import { createApi, createEvent as effectorCreateEvent, createStore as effectorCreateStore, forward, merge, } from 'effector';
import { TAKE_FIRST, TAKE_LAST } from './strategy';
import { cancel, wrap } from './promise';
import { CancelledError, LimitExceededError } from './error';
import { normalizeConfig, normalizeParams } from './config';
export const createReEffectFactory = (createEvent = effectorCreateEvent) => (name, config) => {
    config = normalizeConfig(name, config);
    const instance = createEvent(config);
    const done = createEvent(instance.shortName + '/done');
    const fail = createEvent(instance.shortName + '/fail');
    const anyway = createEvent(instance.shortName + '/finally');
    const cancelled = createEvent(instance.shortName + '/cancelled');
    const stop = createEvent(instance.shortName + '/cancel');
    forward({
        from: merge([done, fail]),
        to: anyway.prepend((payload) => ({
            ...payload,
            status: 'result' in payload ? 'done' : 'fail',
        })),
    });
    let handler = config.handler;
    function use(fn) {
        handler = fn;
        return reeffect;
    }
    use.getCurrent = () => handler;
    const running = new Set();
    const $count = effectorCreateStore(0);
    const count = createApi($count, {
        increment: count => count + 1,
        decrement: count => count - 1,
    });
    const defaultStrategy = config.strategy;
    const defaultLimit = config.limit || Infinity;
    function reeffect(paramsOrStrategy, strategyOrNothing) {
        const { params, strategy } = normalizeParams(defaultStrategy, paramsOrStrategy, strategyOrNothing);
        instance({ params, strategy });
        if (running.size >= defaultLimit) {
            const error = new LimitExceededError(defaultLimit, running.size);
            cancelled({ params, strategy, error });
            return Promise.reject(error);
        }
        if (strategy === TAKE_FIRST && running.size > 0) {
            const error = new CancelledError(strategy);
            cancelled({ params, strategy, error });
            return Promise.reject(error);
        }
        if (strategy === TAKE_LAST && running.size > 0) {
            for (const promise of running) {
                promise[cancel](strategy);
            }
        }
        count.increment();
        let promise;
        let cancelHandlerCallback;
        const onCancel = (callback) => {
            cancelHandlerCallback = callback;
        };
        try {
            const handlerPromise = Promise.resolve(handler(params, onCancel));
            if (cancelHandlerCallback !== undefined) {
                handlerPromise[cancel] = cancelHandlerCallback;
            }
            promise = wrap(handlerPromise);
        }
        catch (error) {
            promise = wrap(Promise.reject(error));
        }
        running.add(promise);
        return promise
            .then(result => {
            if (running.size === 1 && running.has(promise)) {
                done({ params, strategy, result });
            }
            running.delete(promise) && count.decrement();
            return result;
        })
            .catch(error => {
            if (error instanceof CancelledError) {
                cancelled({ params, strategy, error });
            }
            else if (running.size === 1 && running.has(promise)) {
                fail({ params, strategy, error });
            }
            running.delete(promise) && count.decrement();
            throw error;
        });
    }
    stop.watch(() => {
        for (const promise of running) {
            promise[cancel]();
        }
    });
    Object.assign(reeffect, instance);
    reeffect.done = done;
    reeffect.fail = fail;
    reeffect.finally = anyway;
    reeffect.cancelled = cancelled;
    reeffect.cancel = stop;
    reeffect.use = use;
    reeffect.pending = $count.map(count => count > 0);
    reeffect.kind = 'effect';
    return reeffect;
};
