import { createEvent, forward, merge, createStore, createApi } from 'effector';

const TAKE_QUEUE = Symbol('TAKE_QUEUE');
const TAKE_EVERY = Symbol('TAKE_EVERY');
const TAKE_FIRST = Symbol('TAKE_FIRST');
const TAKE_LAST = Symbol('TAKE_LAST');
const isStrategy = (smth) => smth === TAKE_QUEUE ||
    smth === TAKE_EVERY ||
    smth === TAKE_FIRST ||
    smth === TAKE_LAST;
const getStrategyName = (strategy) => String(strategy).replace(/Symbol\((.*)\)/, '$1');

class ReEffectError extends Error {
    constructor(message) {
        super(message);
        delete this.stack;
    }
}
class CancelledError extends ReEffectError {
    constructor(strategy) {
        const name = strategy ? getStrategyName(strategy) : undefined;
        let msg;
        switch (strategy) {
            case TAKE_FIRST:
                msg = `Cancelled due to "${name}" strategy, there are already running effects`;
                break;
            case TAKE_LAST:
                msg = `Cancelled due to "${name}" strategy, new effect was added`;
                break;
            case TAKE_EVERY:
            case TAKE_QUEUE:
                msg = `Hm?.. Cancelled due to "${name}", but should not happen...`;
                break;
            default:
                msg = `Cancelled with "cancel" method, cancel all already running effects`;
                break;
        }
        super(msg);
    }
}
class LimitExceededError extends ReEffectError {
    constructor(limit, running) {
        super(`Cancelled due to limit of "${limit}" exceeded, there are already ${running} running effects`);
    }
}

const cancel = Symbol('cancel');
function wrap(promise) {
    let cancelPromise;
    const cancelable = new Promise((_, reject) => {
        cancelPromise = function (strategy) {
            reject(new CancelledError(strategy));
            if (cancel in promise && typeof promise[cancel] === 'function') {
                promise[cancel]();
            }
        };
    });
    return Object.assign(Promise.race([promise, cancelable]), {
        [cancel]: cancelPromise,
    });
}

function normalizeConfig(name, config) {
    if (config === undefined) {
        config = typeof name === 'object' ? name : {};
    }
    if (typeof name === 'string') {
        config.name = name;
    }
    if (config.handler === undefined) {
        config.handler = noop;
    }
    if (config.strategy === undefined) {
        config.strategy = TAKE_EVERY;
    }
    return config;
}
function normalizeParams(defaultStrategy, paramsOrStrategy, strategyOrNothing) {
    let params;
    let strategy;
    if (isStrategy(paramsOrStrategy)) {
        strategy = paramsOrStrategy;
        params = undefined;
    }
    else {
        strategy = strategyOrNothing || defaultStrategy;
        params = paramsOrStrategy;
    }
    return { params, strategy };
}
function noop() {
    console.error('no handler used');
    return Promise.resolve();
}

const createReEffectFactory = (createEvent$1 = createEvent) => (name, config) => {
    config = normalizeConfig(name, config);
    const instance = createEvent$1(config);
    const done = createEvent$1(instance.shortName + '/done');
    const fail = createEvent$1(instance.shortName + '/fail');
    const anyway = createEvent$1(instance.shortName + '/finally');
    const cancelled = createEvent$1(instance.shortName + '/cancelled');
    const stop = createEvent$1(instance.shortName + '/cancel');
    forward({
        from: merge([done, fail]),
        to: anyway.prepend((payload) => ({
            ...payload,
            status: 'result' in payload ? 'done' : 'fail',
        })),
    });
    let handler = config.handler;
    function use(fn) {
        handler = fn;
        return reeffect;
    }
    use.getCurrent = () => handler;
    const running = new Set();
    const $count = createStore(0);
    const count = createApi($count, {
        increment: count => count + 1,
        decrement: count => count - 1,
    });
    const defaultStrategy = config.strategy;
    const defaultLimit = config.limit || Infinity;
    function reeffect(paramsOrStrategy, strategyOrNothing) {
        const { params, strategy } = normalizeParams(defaultStrategy, paramsOrStrategy, strategyOrNothing);
        instance({ params, strategy });
        if (running.size >= defaultLimit) {
            const error = new LimitExceededError(defaultLimit, running.size);
            cancelled({ params, strategy, error });
            return Promise.reject(error);
        }
        if (strategy === TAKE_FIRST && running.size > 0) {
            const error = new CancelledError(strategy);
            cancelled({ params, strategy, error });
            return Promise.reject(error);
        }
        if (strategy === TAKE_LAST && running.size > 0) {
            for (const promise of running) {
                promise[cancel](strategy);
            }
        }
        count.increment();
        let promise;
        let cancelHandlerCallback;
        const onCancel = (callback) => {
            cancelHandlerCallback = callback;
        };
        try {
            const handlerPromise = Promise.resolve(handler(params, onCancel));
            if (cancelHandlerCallback !== undefined) {
                handlerPromise[cancel] = cancelHandlerCallback;
            }
            promise = wrap(handlerPromise);
        }
        catch (error) {
            promise = wrap(Promise.reject(error));
        }
        running.add(promise);
        return promise
            .then(result => {
            if (running.size === 1 && running.has(promise)) {
                done({ params, strategy, result });
            }
            running.delete(promise) && count.decrement();
            return result;
        })
            .catch(error => {
            if (error instanceof CancelledError) {
                cancelled({ params, strategy, error });
            }
            else if (running.size === 1 && running.has(promise)) {
                fail({ params, strategy, error });
            }
            running.delete(promise) && count.decrement();
            throw error;
        });
    }
    stop.watch(() => {
        for (const promise of running) {
            promise[cancel]();
        }
    });
    Object.assign(reeffect, instance);
    reeffect.done = done;
    reeffect.fail = fail;
    reeffect.finally = anyway;
    reeffect.cancelled = cancelled;
    reeffect.cancel = stop;
    reeffect.use = use;
    reeffect.pending = $count.map(count => count > 0);
    reeffect.kind = 'effect';
    return reeffect;
};

const createReEffect = createReEffectFactory();

export { CancelledError, LimitExceededError, ReEffectError, TAKE_EVERY, TAKE_FIRST, TAKE_LAST, createReEffect, createReEffectFactory };
//# sourceMappingURL=index.js.map
